// Copyright 2023-2025 Hugo Osvaldo Barrera
//
// SPDX-License-Identifier: ISC

use std::{
    ffi::CString,
    fs::File,
    io::Write,
    os::{
        fd::{FromRawFd, OwnedFd, RawFd},
        unix::{ffi::OsStringExt as _, net::UnixListener},
    },
    path::PathBuf,
};

use lexopt::{
    Arg::{Long, Short},
    ValueExt as _,
};
use nix::unistd::pipe;
use wayrs_client::{global::BindError, ConnectError, Connection, IoMode};
use wayrs_protocols::security_context_v1::WpSecurityContextManagerV1;

fn usage() {
    println!(concat!(
        "Usage: way-secure [OPTIONS] --socket-fd SOCKET_FD\n",
        "       way-secure [OPTIONS] --socket-path SOCKET_PATH\n\n",
        "Create Wayland security context at SOCKET_FD\n",
        "\t-e ENGINE    Engine ID for security context\n",
        "\t-a APP_ID    App ID for security context\n",
        "\t-i INSTANCE  Instance ID for security context\n",
        "\t-c FD        Stop listening for new connections when FD closes\n",
        "\t-r FD        Notify readiness on FD\n",
        "\t-h           Show help and exit\n",
        "\nSee 'man way-secure' for details.",
    ));
}

fn main() -> Result<(), Box<dyn std::error::Error>> {
    let mut parser = lexopt::Parser::from_env();

    let mut socket_fd: Option<RawFd> = None;
    let mut socket_path: Option<PathBuf> = None;
    let mut sandbox_engine: Option<CString> = None;
    let mut app_id: Option<CString> = None;
    let mut instance_id: Option<CString> = None;
    let mut close_fd: Option<RawFd> = None;
    let mut ready_fd: Option<RawFd> = None;

    while let Some(arg) = parser.next()? {
        match arg {
            Short('s') | Long("socket-fd") => socket_fd = Some(parser.value()?.parse()?),
            Short('S') | Long("socket-path") => socket_path = Some(parser.value()?.parse()?),
            Short('e') | Long("sandbox-engine") => {
                sandbox_engine = Some(CString::new(parser.value()?.into_vec())?)
            }
            Short('a') | Long("app-id") => app_id = Some(CString::new(parser.value()?.into_vec())?),
            Short('i') | Long("instance-id") => {
                instance_id = Some(CString::new(parser.value()?.into_vec())?)
            }
            Short('c') | Long("close-fd") => close_fd = Some(parser.value()?.parse()?),
            Short('r') | Long("ready-fd") => ready_fd = Some(parser.value()?.parse()?),
            Short('h') | Long("help") => {
                usage();
                std::process::exit(0);
            }
            _ => {
                usage();
                println!();
                return Err(arg.unexpected().into());
            }
        }
    }
    let listen_fd = match (socket_fd, socket_path) {
        // SAFETY: user-supplied fd must be open and suitable for assuming ownership.
        (Some(fd), None) => unsafe { OwnedFd::from_raw_fd(fd) },
        (None, Some(path)) => UnixListener::bind(path)?.into(),
        (None, None) => {
            usage();
            return Err("No socket fd or socket path specified".into());
        }
        (Some(_), Some(_)) => {
            usage();
            return Err("Cannot use both socket fd and socket path.".into());
        }
    };

    // SAFETY: this file descriptor is passed by the parent caller and must be valid.
    let ready_fd = ready_fd.map(|fd| unsafe { OwnedFd::from_raw_fd(fd) });

    let mut alive_fd = None;
    let close_fd = if let Some(close_fd) = close_fd {
        // SAFETY: this file descriptor is passed by the parent caller and must be valid.
        unsafe { OwnedFd::from_raw_fd(close_fd) }
    } else {
        // If close_fd is provided, `way-secure` will exit immediately. If no close_fd is provided,
        // way-secure will continue running in background, and the lifetime of the secure context will
        // be tied to the lifetime of the process.
        let (r, w) = pipe()?;
        alive_fd = Some(w);
        r
    };

    create_context(listen_fd, close_fd, sandbox_engine, app_id, instance_id)?;
    if let Some(fd) = ready_fd {
        File::from(fd).write_all(b"\n")?;
    }

    if alive_fd.is_some() {
        // We hold the other end of the pipe that tells the compositor to continue.
        // Linger indefinitely so the compositor keeps the socket alive.
        loop {
            std::thread::park();
        }
    }

    // Exit immediately if we have a block_fd.
    Ok(())
}

struct State {}

fn create_context(
    listed_fd: OwnedFd,
    close_fd: OwnedFd,
    sandbox_engine: Option<CString>,
    app_id: Option<CString>,
    instance_id: Option<CString>,
) -> Result<(), CreateContextError> {
    let mut conn = Connection::connect().map_err(CreateContextError::Connect)?;
    conn.blocking_roundtrip().map_err(CreateContextError::Io)?;

    let mut state = State {};
    conn.dispatch_events(&mut state);

    let context_manager = conn
        .bind_singleton::<WpSecurityContextManagerV1>(1..=1)
        .map_err(CreateContextError::Bind)?;
    let context = context_manager.create_listener(&mut conn, listed_fd, close_fd);
    if let Some(sandbox_engine) = sandbox_engine {
        context.set_sandbox_engine(&mut conn, sandbox_engine);
    };
    if let Some(app_id) = app_id {
        context.set_app_id(&mut conn, app_id);
    };
    if let Some(instance_id) = instance_id {
        context.set_instance_id(&mut conn, instance_id);
    };
    context.commit(&mut conn);

    conn.flush(IoMode::Blocking)
        .map_err(CreateContextError::Io)?;
    conn.recv_events(IoMode::Blocking)
        .map_err(CreateContextError::Io)?;
    conn.dispatch_events(&mut state);

    Ok(())
}

#[derive(Debug)]
enum CreateContextError {
    Connect(ConnectError),
    Bind(BindError),
    Io(std::io::Error),
}

impl std::fmt::Display for CreateContextError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            CreateContextError::Connect(e) => write!(f, "Connecting to compositor: {e}"),
            CreateContextError::Bind(e) => {
                write!(f, "Binding to wp_security_context_manager_v1: {e}")
            }
            CreateContextError::Io(e) => write!(f, "Compositor Input/Output: {e}"),
        }
    }
}
impl std::error::Error for CreateContextError {}
